/***********************************************************************
tcping.exe -- A tcp probe utility
Copyright (C) 2005 Eli Fulkerson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

----------------------------------------------------------------------

Other license terms may be negotiable.  Contact the author if you would
like a copy that is licensed differently.

Contact information (as well as this program) lives at http://www.elifulkerson.com

----------------------------------------------------------------------

This application includes public domain code from the Winsock Programmer's FAQ:
  http://www.tangentsoft.net/wskfaq/
... and a big thank you to the maintainers and contributers therein.

***********************************************************************/

#include "ws-util.h"

#include <winsock2.h>

#include <iostream>
#include <string.h>
#include <time.h>

#include <sys/timeb.h> // for ftime

#define NUM_PROBES 4

using namespace std;

////////////////////////////////////////////////////////////////////////
// Prototypes

u_long LookupAddress(const char* pcHost);
SOCKET EstablishConnection(u_long nRemoteAddr, u_short nPort, int ping_interval, int force_send_byte);
SOCKET HTTP_EstablishConnection(u_long nRemoteAddr, u_short nPort);
LARGE_INTEGER cpu_frequency;
LARGE_INTEGER response_timer1;
LARGE_INTEGER response_timer2;

LARGE_INTEGER http_timer1;
LARGE_INTEGER http_timer2;



int CTRL_C_ABORT;



const int BufferSize = 1024;

bool SendHttp(SOCKET sd, char* server, char* document, int use_post)
{

	
	char message[1024];
	
	if (use_post == 0) {
		if (document != NULL) {
			sprintf( message, "GET /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\nUser-Agent: httping.exe (www.elifulkerson.com)\r\n\r\n", document, server);
		} else {
			sprintf( message, "GET /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\nUser-Agent: httping.exe (www.elifulkerson.com)\r\n\r\n", "/", server);
		}
	} else {
		if (document != NULL) {
			sprintf( message, "POST /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\nUser-Agent: httping.exe (www.elifulkerson.com)\r\n\r\n", document, server);
		} else {
			sprintf( message, "POST /%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\nUser-Agent: httping.exe (www.elifulkerson.com)\r\n\r\n", "/", server);
			
		}
	}
	
	const int messageLen = strlen(message);
	
	//cout << endl << message << endl;
	
    // Send the string to the server
    if (send(sd, message, messageLen, 0) != SOCKET_ERROR) {
		return true;
	}
	else {
		return false;
    }
}


//// ReadReply /////////////////////////////////////////////////////////
// Read the reply packet and check it for sanity.  Returns -1 on 
// error, 0 on connection closed, > 0 on success.

int ReadReply(SOCKET sd, int &bytes_received, int &http_status)
{
    // Read reply from server
    char acReadBuffer[BufferSize];
    char acTrashBuffer[BufferSize];
    int nTotalBytes = 0;
    int nNewBytes = 0;
    while (1) {
	
		if (nTotalBytes < BufferSize) {
			//cout << "receiving x";
			nNewBytes = recv(sd, acReadBuffer + nTotalBytes, BufferSize - nTotalBytes, 0);
			//cout << "got " << nNewBytes << " new bytes x!" << endl;
		} else {
			//cout << "receiving y";
			nNewBytes = recv(sd, acTrashBuffer, BufferSize, 0);
			//cout << "got " << nNewBytes << " new bytes y!" << endl;
			//cout << acTrashBuffer;
		}
	
        if (nNewBytes == SOCKET_ERROR) {
            return -1;
        }
        else if (nNewBytes == 0) {
            //cerr << "Connection closed by peer." << endl;
            //return 0;
	    	break;
        }

		//cout << nNewBytes << endl;
        nTotalBytes += nNewBytes;
    }
    
    //cout << endl << endl << acReadBuffer << endl;
    bytes_received =  nTotalBytes;
    
    //parse out the http status from the first line of the response    
    char* statusptr;
    char* tmpptr;
    // hop over the initial "HTTP/1.1"
    statusptr = strchr(acReadBuffer, ' ');
    ++statusptr;
    
	tmpptr = strchr(statusptr, ' ');
    // should be at the " " past the error code now
    *tmpptr = '\0';
    
    http_status = atoi(statusptr);
    return 0;
}





void controlc() {
	if (CTRL_C_ABORT == 1) {
		cout.flush();
		cout << "Wow, you really mean it.  I'm sorry... I'll stop now. :(" << endl;
		exit(1);
	}
	cout << "Control-C" << endl;
	CTRL_C_ABORT = 1;
	//return true;
	
}

int DoWinsock( char* pcHost, int nPort, int times_to_ping, int ping_interval, int include_timestamp, int beep_mode, int ping_timeout, int relookup_interval, int auto_exit_on_success, int old_style_time, int force_send_byte, int include_url, int use_http, char* docptr, int use_post)
{
	SetConsoleCtrlHandler((PHANDLER_ROUTINE)&controlc, TRUE);
	CTRL_C_ABORT = 0;
	
	// old_style_time
	int start_time = 0;
	int end_time = 0;
	int start_milltime = 0;
	int end_milltime = 0;
	int total_ms = 0;			// total number of milliseconds per pass
	
	int http_start_time = 0;
	int http_end_time = 0;
	int http_start_milltime = 0;
	int http_end_milltime = 0;
	int http_total_ms = 0;			// total number of milliseconds per pass
	
	int bytes_received = 0;
	int http_status = 0;
	
	float bps = 0;
	
	int number_of_pings = 0;	// total number of tcpings issued
	double running_total_ms = 0;	// running total of values of pings... divide by number_of_pings for avg
	double lowest_ping = 50000;		// lowest ping in series ... starts high so it will drop
	double max_ping = 0;			// highest ping in series
	
	int success_counter = 0;
	int failure_counter = 0;
	int deferred_counter = 0;

	// oldstyle timebuffer
	struct _timeb timebuffer;
	struct _timeb httptimebuffer;

	int loopcounter = 0;			// number of probes to send
	
	// Timestamp variablees
	time_t rawtime;
	struct tm * timeinfo;
	char dateStr [11];
	char timeStr [9];

	int beep_flag = -1;  // 0 for we're down, 1 for we're up
	double response_time;
	double http_response_time;
	
	int have_valid_target = 1;

	cout << endl;

    // Find the server's address
    u_long nRemoteAddress = LookupAddress(pcHost);
    if (nRemoteAddress == INADDR_NONE) {
	cout << endl << "DNS: Could not find host " << pcHost << ", aborting." << endl;
        return 3;
    }
    in_addr Address;
    memcpy(&Address, &nRemoteAddress, sizeof(u_long)); 
   
    
	while ((loopcounter < times_to_ping || times_to_ping == -1) && CTRL_C_ABORT == 0 )
	{
		
		if (((number_of_pings % relookup_interval == 0) && (relookup_interval != -1) && number_of_pings > 0) || have_valid_target == 0) {
			nRemoteAddress = LookupAddress(pcHost);
			if (nRemoteAddress == INADDR_NONE) {
				if (have_valid_target == 1 ) {
					cout << "DNS: Could not find host - " << pcHost << endl;
				}
				have_valid_target = 0;
			} else {
				
				have_valid_target = 1;
				memcpy(&Address, &nRemoteAddress, sizeof(u_long)); 
				cout << "DNS: " << pcHost << " is " << inet_ntoa(Address) << endl;
			}
			
		}
		
		if (include_timestamp == 1) {
			//_strdate( dateStr);
			_strtime( timeStr );
			time ( &rawtime );
			timeinfo = localtime ( &rawtime );
			strftime(dateStr, 11, "%Y:%m:%d",timeinfo);
			cout << dateStr << " " << timeStr << " ";
		}

		
		
		if (have_valid_target == 1) {

		cout << "Probing " << inet_ntoa(Address) << ":" << nPort << "/tcp - ";

		SOCKET sd;
		
		
		if (old_style_time == 0) {
		
			// apparently... QueryPerformanceCounter isn't thread safe unless we do this
			SetThreadAffinityMask(GetCurrentThread(),1); 
		
				// start the timer right before we do the connection
			QueryPerformanceFrequency((LARGE_INTEGER *)&cpu_frequency);
			QueryPerformanceCounter((LARGE_INTEGER *) &response_timer1);

		    // Connect to the server
		    if (use_http == 0) {
		    	sd = EstablishConnection(nRemoteAddress, htons(nPort), ping_timeout, force_send_byte);
	    	} else {
		    	sd = HTTP_EstablishConnection(nRemoteAddress, htons(nPort));
		    	
	    	}

			// grab the timeout as early as possible
			QueryPerformanceCounter((LARGE_INTEGER *) &response_timer2);
		} else {
			// start the timer right before we do the connection
			_ftime(&timebuffer);
			start_milltime = timebuffer.millitm;
			start_time = timebuffer.time;

		    // Connect to the server
		    if (use_http == 0) {
		    	sd = EstablishConnection(nRemoteAddress, htons(nPort), ping_timeout, force_send_byte);
	    	} else {
		    	sd = HTTP_EstablishConnection(nRemoteAddress, htons(nPort));
		    	
	    	}
	    	
			// grab the timeout as early as possible
	   		_ftime(&timebuffer);	
		}
		
	

	    if (sd == INVALID_SOCKET) {
			cout <<  WSAGetLastErrorMessage("");
			failure_counter++;
		    	
		        if (beep_mode == 4 || beep_mode == 1 || (beep_mode == 3 && beep_flag == 1)) {
				cout << " " << char(7) << "*" << char(7) << "*";
			}
			beep_flag = 0;
	    }
		else {
			
			if (use_http == 0) {
				cout << "Port is open";
				success_counter++;
			} else {
				// consider only incrementing if http response @@
				cout << "HTTP is open";
				success_counter++;
				
				// send http send/response
				
				if (old_style_time == 0) {
					SetThreadAffinityMask(GetCurrentThread(),1); 
			
					QueryPerformanceFrequency((LARGE_INTEGER *)&cpu_frequency);
					QueryPerformanceCounter((LARGE_INTEGER *) &http_timer1);
					
					SendHttp(sd,pcHost, docptr, use_post);
					ReadReply(sd, bytes_received, http_status);
					QueryPerformanceCounter((LARGE_INTEGER *) &http_timer2);
	
				} else {
					SendHttp(sd,pcHost,docptr, use_post);
					
					_ftime(&httptimebuffer);
					http_start_milltime = httptimebuffer.millitm;
					http_start_time = httptimebuffer.time;
					
					SendHttp(sd,pcHost, docptr, use_post);
					ReadReply(sd, bytes_received, http_status);
					
					_ftime(&httptimebuffer);
					
				}	
				
			}

			if (beep_mode == 4 || beep_mode == 2 || (beep_mode == 3 && beep_flag == 0)) {
				cout << " *" << char(7);
			}
			beep_flag = 1;
			
  	   }
		    // Shut connection down
	if (ShutdownConnection(sd)) {
		// room here for connection shutdown success check...
	}
		else {
		// room here for connection shutdown failure check...
	}
		

		if (old_style_time == 0) {
	
       		response_time=( (double) ( (response_timer2.QuadPart - response_timer1.QuadPart) * (double) 1000.0 / (double) cpu_frequency.QuadPart) );
		    http_response_time=( (double) ( (http_timer2.QuadPart - http_timer1.QuadPart) * (double) 1000.0 / (double) cpu_frequency.QuadPart) );

		

		} else {
			end_milltime = timebuffer.millitm;
			end_time = timebuffer.time;

			total_ms = ((end_time * 1000) + end_milltime) - ((start_time *1000) + start_milltime);

			response_time = total_ms;
			
			http_end_milltime = httptimebuffer.millitm;
			http_end_time = httptimebuffer.time;

			http_total_ms = ((http_end_time * 1000) + http_end_milltime) - ((http_start_time *1000) + http_start_milltime);

			http_response_time = http_total_ms;
		}

		if (old_style_time == 0) {
			cout.precision(3);
			cout.setf(ios::showpoint);
			cout.setf(ios::fixed);
		}
		cout << " - time=" << response_time << "ms ";

		if (use_http == 1) {
			if (include_url == 1) {
				if (docptr != NULL) {
					cout << "page:" << "http://" << pcHost << "/" << docptr << " ";
				} else {
					cout << "page:" << "http://" << pcHost << " ";
				}
			}
	
			cout << "rcv_time=" << http_response_time << " ";
			cout << "status=" << http_status << " ";
			cout << "bytes=" << bytes_received << " ";
		
			bps = bytes_received * 1000 / http_response_time;
			bps = bps * 8 / 1000;
		
			cout << "kbit/s=~" << bps << " ";
		}
	
		cout << endl;				
		
	
		// Calculate the statistics...
		number_of_pings++;

		if (sd != INVALID_SOCKET) {
			running_total_ms += response_time;

			if (response_time < lowest_ping)
			{
				lowest_ping = response_time;
			}

			if (response_time > max_ping)
			{
				max_ping = response_time;
			}
		}

		loopcounter++;
		if ((loopcounter == times_to_ping) || ((auto_exit_on_success == 1) && (success_counter > 0))) {
			break;
		}
		
	} else {
		// no valid target
		response_time = 0;
		deferred_counter++;
		cout << "No host to ping." << endl;
	}
		
		int zzz = 0;
		int wakeup = (ping_interval * 1000) - response_time;
		if (wakeup > 0 )
		{
			while (zzz < wakeup && CTRL_C_ABORT ==0) {
				Sleep(10);
				zzz += 10;
			}
		}
	}
	
	cout << endl << "Ping statistics for " << inet_ntoa(Address) << ":" << nPort << endl;
	cout << "     " << number_of_pings << " probes sent. "<< endl;
	cout << "     " << success_counter << " successful, " << failure_counter << " failed." << endl;
	if (deferred_counter > 0) {
		cout << "     " << deferred_counter << " skipped due to failed DNS lookup." << endl;
	}
	if (success_counter > 0) {
		if ( failure_counter > 0) {
			cout << "Approximate trip times in milli-seconds (successful connections only):" << endl;
		} else {
			cout << "Approximate trip times in milli-seconds:" << endl;
		}
		cout.precision(3);
		cout << "     Minimum = " << lowest_ping << "ms, Maximum = " << max_ping << "ms, Average = " << running_total_ms/number_of_pings << "ms" <<endl;
	} else {
		cout << "Was unable to connect, cannot provide trip statistics." << endl;
	}
		
	// report our total, abject failure.
	if (success_counter == 0) {
		return 1;
	}
	
	// return our intermittent failure
	if (success_counter > 0 && failure_counter > 0) {
		return 2;	
	}
		
    return 0;
}


//// LookupAddress /////////////////////////////////////////////////////
// Given an address string, determine if it's a dotted-quad IP address
// or a domain address.  If the latter, ask DNS to resolve it.  In
// either case, return resolved IP address.  If we fail, we return
// INADDR_NONE.

u_long LookupAddress(const char* pcHost)
{
    u_long nRemoteAddr = inet_addr(pcHost);
    if (nRemoteAddr == INADDR_NONE) {
        // pcHost isn't a dotted IP, so resolve it through DNS
        hostent* pHE = gethostbyname(pcHost);
        if (pHE == 0) {
            return INADDR_NONE;
        }
        nRemoteAddr = *((u_long*)pHE->h_addr_list[0]);
    }

    return nRemoteAddr;
}


//// EstablishConnection ///////////////////////////////////////////////
// Connects to a given address, on a given port, both of which must be
// in network byte order.  Returns newly-connected socket if we succeed,
// or INVALID_SOCKET if we fail.

SOCKET EstablishConnection(u_long nRemoteAddr, u_short nPort, int ping_timeout, int force_send_byte)
{
	
	LARGE_INTEGER timer1;
	LARGE_INTEGER timer2;
	LARGE_INTEGER cpu_freq;
	
	double time_so_far;

	// Create a stream socket
	SOCKET sd = socket(AF_INET, SOCK_STREAM, 0);

	//-------------------------
	// Set the socket I/O mode: In this case FIONBIO
	// enables or disables the blocking mode for the 
	// socket based on the numerical value of iMode.
	// If iMode = 0, blocking is enabled; 
	// If iMode != 0, non-blocking mode is enabled.
	u_long iMode = 1;
	ioctlsocket(sd, FIONBIO, &iMode);
		
        sockaddr_in sinRemote;
        sinRemote.sin_family = AF_INET;
        sinRemote.sin_addr.s_addr = nRemoteAddr;
        sinRemote.sin_port = nPort;

        QueryPerformanceFrequency((LARGE_INTEGER *)&cpu_freq);
        QueryPerformanceCounter((LARGE_INTEGER *) &timer1);
	
	connect(sd, (sockaddr*)&sinRemote, sizeof(sockaddr_in));
	    
	char *sendy = ".";
	int size = 1;
	int sendstatus = 1000;
	
	bool done = false;
	
	//for (int i=0;i<ping_timeout;i+=10) {
	while (!done && !CTRL_C_ABORT) {
		
		if (force_send_byte == 0) {
			sendstatus = send(sd, sendy, 0, 0);   // should return 0 below
		} else {
			sendstatus = send(sd, sendy, size, 0);   // should return sizeof(sendy) below
		}
		

		// one error code is if you send a send of size 0, the other is if you actually send data.
		if (sendstatus == size && force_send_byte == 1) {
			return sd;
		}
		
		if (sendstatus == 0 && force_send_byte == 0) {
			return sd;
		}
		
		QueryPerformanceCounter((LARGE_INTEGER *) &timer2);
		
		time_so_far = ( (double) ( (timer2.QuadPart - timer1.QuadPart) * (double) 1000.0 / (double) cpu_freq.QuadPart) );
		
		if (time_so_far >= ping_timeout) {
			done = true;
		} else {		
			Sleep(1);
		}
	}

	return INVALID_SOCKET;
    //return sd;
}

SOCKET HTTP_EstablishConnection(u_long nRemoteAddr, u_short nPort)
{

    // Create a stream socket
    SOCKET sd = socket(AF_INET, SOCK_STREAM, 0);

    // would be nice to limit the huge timeout in cases where the tcp connection times out (as opposed
    // to bouncing off a closed port, but this stuff doesn't work for some reason.
    /*int timeout = 10;
    int tosize = sizeof(timeout);
    setsockopt(sd, SOL_SOCKET, SO_SNDTIMEO , (char*)&timeout, (int)&tosize);
    setsockopt(sd, SOL_SOCKET, SO_RCVTIMEO , (char*)&timeout, (int)&tosize);
    */
	
    if (sd != INVALID_SOCKET) {
        sockaddr_in sinRemote;
        sinRemote.sin_family = AF_INET;
        sinRemote.sin_addr.s_addr = nRemoteAddr;
        sinRemote.sin_port = nPort;
        if (connect(sd, (sockaddr*)&sinRemote, sizeof(sockaddr_in)) ==
                SOCKET_ERROR) {
            sd = INVALID_SOCKET;
        }
    }

    return sd;
}